"""Utility functions for prompting the user related to enums."""
from enum import Enum
from typing import Any, Iterable, Type, TypeVar, Union, get_args

import typer

E = TypeVar("E", bound=Enum)


def is_enum_or_enum_union(field_type: Type[Any]) -> bool:
    """Check if the field_type is an enum or a union of enums."""
    # Check if the field_type is an enum
    try:
        if issubclass(field_type, Enum):
            return True
    except TypeError:
        return is_enum_union(field_type)


def is_enum_union(field_type: Any) -> bool:  # noqa: ANN401
    """Check if the field_type is a union of enums.

    :param field_type: The type to check.
    :return: True if field_type is a union of enum subclasses, False otherwise.
    """
    if hasattr(field_type, "__origin__") and field_type.__origin__ is Union:
        try:
            return all((issubclass(arg, Enum) for arg in get_args(field_type)))
        except TypeError:
            pass
    return False


def select_enum_from_enum_union(union_enum_type: Union[Enum]) -> Type[Enum]:
    """Select an enum from a union of enums.

    :param union_enum_type: The union of enums to select from.
    """
    # Check if the field_type is a union of enums
    # Prompt the user to select an enum from the union
    enum_type_options = get_args(union_enum_type)
    selection_string = _format_enum_type_selection(enum_type_options)
    print(selection_string)
    selected_value = typer.prompt(
        f"Please enter a value between 0 and {len(enum_type_options) - 1}: ", type=int
    )
    # Return the selected enum type
    return enum_type_options[selected_value]


def prompt_for_enum_selection(
    enum_class: Union[Enum, Union[Enum]],
    allow_none: bool = False,
) -> Enum:
    """Prompt the user to select an enum member from an enum.

    If provided with a union of enums, prompt the user to select an enum from the union.

    :param enum_class: The enum to select from.
    """
    # Determine the type of enum (int or str) and generate the appropriate prompt
    if is_enum_union(enum_class):
        enum_class = select_enum_from_enum_union(enum_class)

    enum_type = int if issubclass(enum_class, int) else str

    # Prompt the user for a selection
    selection_string = _format_enum_selection(enum_class, enum_type)
    print(selection_string)

    selected_value = typer.prompt(
        f"Please enter a value between 0 and {len(enum_class) - 1}: ", type=int
    )

    return _enum_member_by_index(enum_class, selected_value)


def _format_enum_selection(enum_class: Type[Enum], enum_type: Union[str, int]) -> str:
    """Format the enum selection prompt string."""
    enum_name = format_enum_name(enum_class)
    strs = [f"\nAvailable Geneset {enum_name} Types:"]
    for index, member in enumerate(enum_class):
        display_name = member.name.capitalize() + " "
        display_value = f"{member.value} -- {index}" if enum_type is str else index
        strs.append(f"    {display_name.ljust(30, '-')} {display_value}")
    return "\n".join(strs)


def _format_enum_type_selection(enums: Iterable[Type[Enum]]) -> str:
    """Format the enum type selection prompt string.

    :param enums: The enum types to select from.
    :return: The formatted enum type selection prompt string.
    """
    strs = ["\nPlease Select one of: "]
    for index, enum_class in enumerate(enums):
        enum_name = format_enum_name(enum_class)
        strs.append(f"    {enum_name.ljust(30, '-')} {index}")
    return "\n".join(strs)


def format_enum_name(enum_class: Type[Enum]) -> str:
    """Return the name of the enum class without the 'Geneset' or 'Type' suffix.

    :param enum_class (Type[Enum]): The class of the enumeration.
    :return (str): The name of the enum class without the 'Geneset' or 'Type' suffix.
    """
    return (
        enum_class.__name__.replace("Geneset", "")
        .replace("Type", "")
        .replace("Enum", "")
        .strip()
    )


def _enum_member_by_index(enum_class: Type[E], index: int) -> E:
    """Get the enum member at the specified index from the given enum class.

    :param enum_class (Type[E]): The class of the enumeration.
    :param index (int): The index of the enum member to return.
    :return (E): The enum member at the specified index.
    :raises IndexError: If the index is out of range.
    """
    return list(enum_class)[index]



"""CLI for parsing utility functions."""
from pathlib import Path
from typing import Optional

import typer
from geneweaver.client.parser import general
from geneweaver.client.utils.cli.decorators.errors import print_value_errors
from geneweaver.client.utils.cli.print.file_info import (
    print_metadata_csv,
    print_metadata_xlsx,
)
from geneweaver.client.utils.cli.print.tables import print_tabular_data
from geneweaver.core.parse import csv, xlsx
from geneweaver.core.parse.exceptions import EmptyFileError, UnsupportedFileTypeError
from geneweaver.core.parse.utils import get_file_type
from rich import print
from rich.console import Console

cli = typer.Typer()
console = Console()


@cli.command()
@print_value_errors
def infer_file_format(file_path: Path) -> None:
    """Infer the file format of a data file."""
    file_type = get_file_type(file_path)
    print(f"{file_type.name} - {file_type.value}")


@cli.command()
@print_value_errors
def get_headers(file_path: Path, sheet: str = None) -> None:
    """Get the headers from a data file."""
    try:
        headers, _ = general.get_headers(file_path, sheet_name=sheet)
    except (EmptyFileError, ValueError, UnsupportedFileTypeError) as e:
        print(e)
        raise typer.Exit(code=1) from e

    for idx, header in enumerate(headers):
        print(f"{idx} - {header}")


@cli.command()
@print_value_errors
def get_sheet_names(file_path: Path) -> None:
    """Get the sheet names from an Excel file."""
    try:
        names = xlsx.get_sheet_names(file_path)
    except ValueError as e:
        print(e)
        raise typer.Exit(code=1) from e

    for idx, name in enumerate(names):
        print(f"{idx} - {name}")


@cli.command()
@print_value_errors
def get_metadata(file_path: Path, sheet: Optional[str] = None) -> None:
    """Get the metadata from a data file."""
    file_type = get_file_type(file_path)

    if file_type == "csv":
        header, header_idx = csv.get_headers(file_path)
        try:
            metadata = general.read_metadata(file_path, header_idx)
        except (EmptyFileError, ValueError, UnsupportedFileTypeError) as e:
            print(e)
            raise typer.Exit(code=1) from e

        print_metadata_csv(file_path, metadata)
    elif file_type == "xlsx":
        sheet_names, sheet_metadata, _, _, n_sheets = xlsx.get_metadata(
            file_path, sheet
        )
        print_metadata_xlsx(file_path, sheet_names, sheet_metadata, n_sheets)


@cli.command()
@print_value_errors
def preview(
    file_path: Path, rows_to_read: int = 5, sheet: str = None, prompt: bool = True
) -> None:
    """Preview the data in a data file."""
    file_type = get_file_type(file_path)
    if file_type == "csv":
        _preview_csv(file_path, rows_to_read, prompt)
    elif file_type == "xlsx":
        _preview_xlsx(file_path, rows_to_read, sheet, prompt)


def _preview_csv(file_path: Path, rows_to_read: int = 5, prompt: bool = True) -> None:
    """Preview the data in a CSV file."""
    header, header_idx = csv.get_headers(file_path)
    try:
        metadata = general.read_metadata(file_path, header_idx)
    except (EmptyFileError, ValueError, UnsupportedFileTypeError) as e:
        print(e)
        raise typer.Exit(code=1) from e

    print_metadata_csv(file_path, metadata)

    if prompt:
        typer.confirm("Do you want to preview data?", default=True, abort=True)

    rows = csv.read_to_dict_n_rows(file_path, rows_to_read, header_idx)
    print_tabular_data(header, rows)


def _preview_xlsx(
    file_path: Path, rows_to_read: int = 5, sheet: str = None, prompt: bool = True
) -> None:
    sheet_names, sheet_metadata, headers, headers_idx, n_sheets = xlsx.get_metadata(
        file_path, sheet
    )

    if not sheet:
        print_metadata_xlsx(file_path, sheet_names, sheet_metadata, n_sheets)

        if prompt:
            typer.confirm("Do you want to preview data?", default=True, abort=True)

    zipped_data = zip(sheet_names, sheet_metadata, headers, headers_idx)  # noqa: B905

    for idx, (sheet_name, metadata, header, header_idx) in enumerate(zipped_data):
        print(f"Data for sheet - {sheet_name} - {metadata}")
        rows = xlsx.read_to_dict_n_rows(
            str(file_path), rows_to_read, header_idx, sheet_name
        )
        print_tabular_data(header, rows)

        if prompt and idx < n_sheets - 1:
            typer.confirm("Next sheet?", default=True, abort=True)


def _preview_data(file_path: Path, rows_to_read: int = 5, sheet: str = None) -> None:
    try:
        headers, headers_idx = general.get_headers(file_path, sheet_name=sheet)
        rows = general.data_file_to_dict_n_rows(
            file_path, rows_to_read, headers_idx, sheet_name=sheet
        )
    except ValueError as e:
        print("Something went wrong, the file is probably empty.\n" + str(e))
        raise typer.Exit(code=1) from e
    except (EmptyFileError, UnsupportedFileTypeError) as e:
        print(e)
        raise typer.Exit(code=1) from e

    print_tabular_data(headers, rows)



"""Utility functions for prompting the user to enter a list of values."""
from typing import List, Optional, TypeVar, get_args, get_origin

import typer

T = TypeVar("T")


def prompt_for_list_selection(field_type: T, allow_none: bool = False) -> List[T]:
    """Prompt the user to enter a list of values.

    :param field_type: A list of strings or integers, treated as strings if not
                       integers.
    :return: The list of values entered by the user.
    """
    origin_type = get_origin(field_type)
    if origin_type != list:
        raise TypeError("Expected a typing.List type")

    # Extract the type of the list's elements
    element_type = get_args(field_type)[0]

    # Determine if the elements are integers or strings
    is_int_type = element_type is int

    items = []
    while True:
        item = typer.prompt(
            f"Enter a {'number' if is_int_type else 'string'} "
            "(or just enter to finish): ",
            type=element_type,
            default="",
        )
        if not item:
            break
        items.append(item)
    return items


def is_list_of_str_or_int(field_type: T) -> bool:
    """Check if the field_type is a list of strings or integers.

    :param field_type: The field type to check.
    :return: True if the field type is a list of strings or integers, False otherwise.
    """
    # Check if the field_type is a list
    if get_origin(field_type) is list:
        # Get the type of elements in the list
        element_type = get_args(field_type)[0]
        # Check if the element type is str or int
        return element_type in [str, int]
    return False


def prompt_if_list_contains_duplicates(
    list_to_check: list, message: Optional[str]
) -> None:
    """Prompt the user if the given list contains duplicates.

    :param list_to_check: The list to check for duplicates.
    :param message: The message to display to the user.
    """
    if len(list_to_check) != len(set(list_to_check)):
        if message is None:
            message = "The list contains duplicates. Please try again."
        typer.echo(message)



"""Root of the alpha api subcommand."""
import typer
from geneweaver.client.cli.alpha.api import genes, genesets

HELP_MESSAGE = """
Tools and utilities for interacting with the Geneweaver API.
"""

cli = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
cli.add_typer(genesets.cli, name="genesets", help=genesets.HELP_MESSAGE)
cli.add_typer(genes.cli, name="genes", help=genes.HELP_MESSAGE)



"""Authentication CLI commands."""
import typer
from geneweaver.client import auth
from geneweaver.client.auth import get_access_token, get_id_token
from geneweaver.client.exceptions import AuthenticationError

cli = typer.Typer()

HELP_MESSAGE = """
The auth commands allow you to authenticate with the GeneWeaver API.
"""


@cli.command(name="login")
def _login(reauth: bool = typer.Option(False, "--reauth")) -> None:  # noqa: B008
    """Run the device authorization flow.

    :param reauth: Force a re-authentication
    """
    if not reauth and get_access_token():
        print("You are already logged in")
        raise typer.Exit(code=1)

    try:
        auth.login()
    except AuthenticationError as e:
        print(e)
        raise typer.Exit(code=1) from e


@cli.command()
def print_access_token() -> None:
    """Print the current user's access token."""
    print(get_access_token())


@cli.command()
def print_identity_token() -> None:
    """Print the current user's identity token."""
    print(get_id_token())



"""Functions that wrap the GeneWeaver API on /genesets endpoints."""
from typing import Optional

from geneweaver.client.api.utils import format_endpoint, sessionmanager
from geneweaver.core.enum import GeneIdentifier

ENDPOINT = "genesets"


def get(
    access_token: str, geneset_id: int, gene_id_type: Optional[GeneIdentifier] = None
) -> dict:
    """Get a Geneset by ID.

    :param access_token: User access token
    :param geneset_id: Geneset ID (without the "GS" prefix).
    :param gene_id_type: Gene ID type (one of GeneIdentifier).

    :return: Geneset dict.
    """
    params = {} if gene_id_type is None else {"gene_id_type": int(gene_id_type)}
    with sessionmanager(token=access_token) as session:
        resp = session.get(
            format_endpoint(ENDPOINT, str(geneset_id)),
            params=params,
        )
    return resp.json()


def get_genesets(access_token: str) -> list:
    """Get all visible genesets.

    :param access_token: User access token

    :return: List of genesets dicts.
    """
    with sessionmanager(token=access_token) as session:
        resp = session.get(format_endpoint(ENDPOINT))
    return resp.json()



"""Dataset definition for NCI60's: DNA__Combined_aCGH_gene_summary.xls."""
import pandas as pd
from geneweaver.client.datasets.base import BaseDataset


class DNACombinedaCGHGeneSummary(BaseDataset):
    """Dataset definition for NCI60's: DNA__Combined_aCGH_gene_summary.xls."""

    URL = (
        "https://discover.nci.nih.gov/cellminer/download/processeddataset/"
        "nci60_DNA__Combined_aCGH_gene_summary.zip"
    )
    DS_FOLDER = "nci60_DNA__Combined_aCGH_gene_summary"
    UNZIPPED_LOC = "output/DNA__Combined_aCGH_gene_summary.xls"
    LINKOUT = "https://discover.nci.nih.gov/cellminer/loadDownload.do"

    def __init__(self, base_folder: str = "data") -> None:
        """Initialize the DNACombinedaCGHGeneSummary dataset."""
        super().__init__(base_folder)
        self.download_zip_file()
        self.dataset_skip_rows = 10
        self._pandas_read_f = pd.read_excel

    @property
    def entrez_ids(self) -> pd.DataFrame:
        """Return the dataset's entrez ids."""
        return self.as_pandas().iloc[:, 1]

    @property
    def gene_names(self) -> pd.DataFrame:
        """Return the dataset's gene names."""
        return self.as_pandas().iloc[:, 0]

    @property
    def intensity(self) -> pd.DataFrame:
        """Return the dataset's intensity values."""
        intensity = self.as_pandas().iloc[:, 6:]
        intensity = intensity.replace("-", 0)
        return intensity

    @property
    def intensity_type(self) -> str:
        """Return the dataset's intensity type."""
        return (
            "Average log2 intensity values of the ratio of the cell line "
            "DNA with respect to normal DNA."
        )



"""Utility functions for adding CLI Options."""
# ruff: noqa: ANN001, ANN002, ANN003, ANN201, ANN202
import functools
from inspect import Parameter, signature
from typing import Type

import typer
from pydantic import BaseModel


def add_options_from_model(model: Type[BaseModel], option_prefix: str = ""):
    """Add options to a function from a pydantic model.

    :param model: The pydantic model to add options from.
    :param option_prefix: A prefix to add to the option names.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            model_instance = model(**kwargs)
            return func(model_instance)

        sig = signature(func)
        new_parameters = list(sig.parameters.values())

        for field_name, field in model.__fields__.items():
            option_help = field.field_info.description or field_name
            option_name = (
                f"{option_prefix}_{field_name}" if option_prefix else field_name
            )
            new_param = Parameter(
                option_name,
                Parameter.KEYWORD_ONLY,
                default=typer.Option(..., help=option_help),
            )
            new_parameters.append(new_param)

        wrapper.__signature__ = sig.replace(parameters=new_parameters)
        return wrapper

    return decorator



"""Root command for the parse cli subcommand."""
import typer
from geneweaver.client.cli.alpha.parse import utils

from .convert import convert

HELP_MESSAGE = """
Tools and utilities to parse data files for use in Geneweaver.

The parse commands help to transform data files in various formats into data files that
can be uploaded to Geneweaver.
"""

cli = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")

cli.add_typer(utils.cli, name="utils")

cli.command()(convert)
cli.command(name="cn", help="Alias for `convert` command.")(convert)



"""Convert data files."""
from enum import Enum
from pathlib import Path
from typing import List, Optional

import typer
from geneweaver.client.utils.cli.prompt.generic import (
    prompt_if_none,
    prompt_if_none_or_ask_to_keep,
)
from geneweaver.client.utils.cli.prompt.list import prompt_if_list_contains_duplicates
from geneweaver.client.utils.cli.prompt.pydantic import prompt_for_missing_fields
from geneweaver.core.parse import csv, xlsx
from geneweaver.core.parse.utils import get_file_type
from geneweaver.core.render.batch import format_batch_file
from geneweaver.core.render.csv import format_csv_file
from geneweaver.core.schema.batch import BatchUploadGeneset
from geneweaver.core.schema.gene import GeneValue
from pydantic import ValidationError
from rich.progress import Progress, SpinnerColumn, TextColumn


class CovertFileType(str, Enum):
    """Enum for file types."""

    CSV = "csv"
    BATCH = "gw"


def convert(
    file_path: Path,
    # value_header: Annotated[str, typer.Option(prompt=True)],
    # id_header: Annotated[str, typer.Option(prompt=True)],
    value_header: Optional[str] = None,
    id_header: Optional[str] = None,
    to: CovertFileType = CovertFileType.BATCH,
) -> None:
    """Convert files from one format to another.

    :param file_path: The path to the file to convert.
    :param value_header: The name of the column containing the gene values.
    :param id_header: The name of the column containing the gene IDs.
    :param to: The file type to convert to.

    :raises ValueError: If the file type is not supported.
    """
    file_type = get_file_type(file_path)
    out_file = file_path.with_suffix(f".{to.value}")
    print(f"Converting {file_path} to {out_file}")

    genesets = []

    if file_type == "xlsx":
        genesets = _convert_excel(file_path, id_header, value_header)
    elif file_type == "csv":
        genesets = _covert_csv(file_path, id_header, value_header)

    if to == CovertFileType.CSV:
        for geneset in genesets:
            this_out_file = out_file.with_name(
                f"{out_file}_{geneset.abbreviation}.{to.value}"
            )
            with open(this_out_file, "w") as f:
                f.write(format_csv_file(geneset))
            print(f"Converted {file_path} to {this_out_file}")
    else:
        with open(out_file, "w") as f:
            f.write(format_batch_file(genesets))

        print(f"Converted {file_path} to {out_file}")


def _convert_excel(
    file_path: Path, id_header: Optional[str] = None, value_header: Optional[str] = None
) -> List[BatchUploadGeneset]:
    """Convert an Excel file to a batch file.

    :param file_path: The path to the Excel file.
    :param id_header: The name of the column containing the gene IDs.
    :param value_header: The name of the column containing the gene values.
    """
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Loading document...", total=None)
        data = _parse_excel(file_path)

    file_name = file_path.name.split(".")[0]
    genesets = []

    for sheet_name, _header, sheet_metadata, sheet_data in data:
        print("Working on sheet:", sheet_name)
        gs_name = f"{file_name} - {sheet_name}"
        gs_abbreviation = gs_name.replace(" ", "").replace("-", "_").capitalize()
        gs_description = gs_name + " " + ", ".join(sheet_metadata)
        id_header = prompt_if_none_or_ask_to_keep("ID Header", id_header)
        value_header = prompt_if_none_or_ask_to_keep("Value Header", value_header)
        geneset = _build_geneset(
            name=gs_name, abbreviation=gs_abbreviation, description=gs_description
        )
        geneset.values = _parse_gene_list(  # noqa: PD011
            sheet_data, id_header, value_header
        )
        genesets.append(geneset)

    return genesets


def _covert_csv(
    file_path: Path, id_header: Optional[str] = None, value_header: Optional[str] = None
) -> List[BatchUploadGeneset]:
    """Convert a CSV file to a batch file.

    :param file_path: The path to the CSV file.
    :param id_header: The name of the column containing the gene IDs.
    :param value_header: The name of the column containing the gene values.
    """
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Loading document...", total=None)
        headers, header_idx = csv.get_headers(file_path)
        if header_idx == -1:
            typer.echo("Could not find header row. Aborting.")
            typer.Exit()

        prompt_if_list_contains_duplicates(
            headers,
            "WARNING: Possible duplicate headers."
            "Please use CAUTION. Consider renaming headers and trying again.",
        )

        data = csv.read_to_dict(file_path, header_idx)

        file_name = file_path.name.split(".")[0]

        gs_name = file_name
        gs_abbreviation = gs_name.replace(" ", "").replace("-", "_").capitalize()
        gs_description = gs_name
    id_header = prompt_if_none("id_header", id_header)
    value_header = prompt_if_none("value_header", value_header)
    geneset = _build_geneset(
        name=gs_name, abbreviation=gs_abbreviation, description=gs_description
    )
    geneset.values = _parse_gene_list(data, id_header, value_header)  # noqa: PD011

    return [geneset]


def _parse_excel(file_path: Path) -> List[tuple]:
    """Parse an Excel file.

    :param file_path: The file path to the Excel file.
    :returns: A list of tuples containing the sheet name, headers, metadata, and data.
    """
    sheet_names = xlsx.get_sheet_names(file_path)

    headers, headers_idx = [], []
    for s in sheet_names:
        h, h_idx = xlsx.get_headers(file_path, sheet_name=s)
        prompt_if_list_contains_duplicates(
            h,
            f"WARNING: Possible duplicate headers on sheet {s}. "
            f"Please use CAUTION. Consider renaming headers and trying again.",
        )
        if h_idx == -1:
            typer.echo(f"WARNING: Could not find header row for sheet {s}. Skipping.")
        else:
            headers.append(h)
            headers_idx.append(h_idx)

    sheet_metadata = [
        xlsx.read_metadata(file_path, header_idx, sheet_name=sheet)
        for sheet, header_idx in zip(sheet_names, headers_idx)  # noqa: B905
    ]

    data = [
        xlsx.read_to_dict(file_path, header_idx, sheet_name=sheet)
        for sheet, header_idx in zip(sheet_names, headers_idx)  # noqa: B905
    ]

    return zip(sheet_names, headers, sheet_metadata, data)  # noqa: B905


def _parse_gene_list(
    data: List[dict], id_header: str, value_header: str
) -> List[GeneValue]:
    """Parse a list of genes from a data file.

    :param data: A list of dictionaries representing the data from a data file.
    :param id_header: The name of the column containing the gene IDs.
    :param value_header: The name of the column containing the gene values.
    :returns: A list of GeneValue instances.
    """
    gene_list = []

    for row in data:
        try:
            j = GeneValue(symbol=row[id_header], value=row[value_header])
            gene_list.append(j)
        except ValidationError:
            continue

    return gene_list


def _build_geneset(**kwargs: str) -> BatchUploadGeneset:
    """Build a geneset from kwargs and by prompting for missing fields.

    :param kwargs: Keyword arguments to pass to the BatchUploadGeneset constructor.
    :returns: A BatchUploadGeneset instance.
    """
    geneset_args = prompt_for_missing_fields(
        BatchUploadGeneset, kwargs, exclude={"values"}
    )

    batch_upload = BatchUploadGeneset(values=[], **geneset_args)

    return batch_upload



"""GeneWeaver Client configuration module."""
from typing import Optional, Type

from pydantic import BaseSettings, validator


class Settings(BaseSettings):
    """Settings class for GeneWeaver Client."""

    AUTH_DOMAIN = "geneweaver.auth0.com"
    AUTH_CLIENT_ID = "tvcvnbBJQr15jFds8ZkwgHoqRJ0IuGJC"
    AUTH_ALGORITHMS = ["RS256"]
    AUTH_SCOPES = ["openid", "profile", "email"]

    API_HOST: str = "https://geneweaver-prod.jax.org"
    API_PATH: str = "/api"

    API_URL: Optional[str] = None

    @validator("API_URL")
    def validate_api_url(
        cls: Type["Settings"], v: Optional[str], values: dict  # noqa: N805
    ) -> str:
        """Construct the API URL if not explicitly set."""
        if not v:
            return values["API_HOST"] + values["API_PATH"]
        return v

    API_KEY: Optional[str] = None

    class Config:
        """Settings configuration."""

        env_file = ".env"
        env_prefix = "GW_CLIENT_"



"""Root of the beta cli subcommand."""
import typer
from geneweaver.client.cli.beta import auth

HELP_MESSAGE = """
These commands are in beta testing.

They are subject to future change and/or removal. Beta commands are
[italic]intended[/italic] to be released beyond beta testing, but may have bugs or other
issues. There is also [bold]no guarantee[/bold] that beta commands will be released
beyond beta testing.

:warning: [bold red]Use at your own risk.[/bold red] :warning:
"""

cli = typer.Typer(no_args_is_help=True)

cli.add_typer(auth.cli, name="auth", help=auth.HELP_MESSAGE)



"""API related utilities, helpers, and other internal functions."""
from contextlib import contextmanager
from typing import Any, Optional

import requests
from geneweaver.client.api.exc import GeneweaverAPIError
from geneweaver.client.core.config import settings


def _raise_for_status_hook(
    response: requests.Response, *args: Any, **kwargs: Any  # noqa: ANN401
) -> None:
    """Have request responses raise an exception for non-200 status codes."""
    response.raise_for_status()


@contextmanager
def sessionmanager(token: Optional[str] = None) -> requests.Session:
    """Context manager for a requests.Session object.

    This context manager will do everything that a requests.Session context manager
    does, but will also raise an exception if the response status code is not 200.
    It will also wrap all exceptions inheriting from
    `requests.exceptions.RequestException` in a GeneweaverAPIException.
    """
    with requests.Session() as session:
        session.hooks = {"response": _raise_for_status_hook}
        if token is not None:
            session.headers.update({"Authorization": f"Bearer {token}"})
        try:
            yield session
        except requests.exceptions.RequestException as err:
            # TODO: We SHOULD try extracting the error message from the response,
            #  could even check the JSON.
            err_str = (
                f"There was a problem calling the Geneweaver API: {err.response.text}"
            )
            raise GeneweaverAPIError(err_str) from err
        finally:
            session.close()


def format_endpoint(directory: str, *args: str) -> str:
    """Format the endpoint with the API host and path."""
    return "/".join([settings.API_URL, directory] + list(args))



"""A module for parsing batch files."""
import csv
from pathlib import Path
from typing import Dict, List, Optional

from geneweaver.core.parse import batch
from geneweaver.core.schema.batch import BatchUploadGeneset


def to_csv(
    batch_file: Path,
    output_directory: Optional[Path] = None,
    prefix: Optional[str] = None,
    geneset_ids: Optional[List[str]] = None,
    read_file: bool = True,
) -> Optional[List[Path]]:
    """Convert a batch file to a CSV file.

    :param batch_file: The path to the batch file.
    :param output_directory: The directory to write the CSV file to. If not provided,
    :param prefix: A prefix to add to the CSV file name.
    :param geneset_ids: A list of geneset IDs to use for the CSV file names.
    the CSV file will be written to the current working directory.
    """
    if read_file:
        with open(batch_file, "r") as f:
            contents = f.read()
    else:
        contents = batch_file

    genesets = batch.process_lines(contents)

    csv_files = [
        write_geneset_to_csv(geneset, None, output_directory, prefix, gs_id)
        for geneset, gs_id in zip(genesets, geneset_ids)  # noqa: B905
    ]

    return csv_files


def to_csv_indexed(
    batch_file: Path, index_file: Path, output_directory: Optional[Path] = None
) -> List[Path]:
    """Convert a batch file to a CSV file, using the index file to name the CSV files.

    :param batch_file: The path to the batch file.
    :param index_file: The path to the index file.
    :param output_directory: The directory to write the CSV file to. If not provided,
    the CSV file will be written to the current working directory.
    :return: A list of the paths to the created CSV files.
    """
    index_data = read_index_file(index_file)

    with open(batch_file, "r") as f:
        contents = f.read()

    genesets = batch.process_lines(contents)

    filenames = []

    for geneset in genesets:
        try:
            index = index_data["GW Name"].index(geneset.name)
            disease = index_data["disease name"][index].lower().replace(" ", "_")
            gs_id = index_data["GW gene set id"][index]
            uberon_id = index_data["UBERON id"][index]
            filename = write_geneset_to_csv(
                geneset, uberon_id, output_directory, disease, gs_id
            )
            filenames.append(filename)
        except ValueError:
            print(f"Could not find geneset {geneset.name} in index file")
            continue

    return filenames


def read_index_file(index_file: Path) -> Dict[str, List[str]]:
    """Read the index file and return a dictionary of lists.

    :param index_file: The path to the index file.
    :return: A dictionary of lists.
    """
    with open(index_file, "r", errors="replace") as tsvfile:
        header = next(tsvfile).strip().split("\t")
        data_dict = {col: [] for col in header}

    # Read the TSV file and populate the lists in the dictionary
    with open(index_file, "r", errors="replace") as tsvfile:
        reader = csv.DictReader(tsvfile, delimiter="\t")

        for row in reader:
            for col in header:
                data_dict[col].append(row[col])

    return data_dict


def _index_of_geneset(name: str, index_data: Dict[str, List[str]]) -> int:
    return index_data["name"].index(name)


def write_geneset_to_csv(
    geneset: BatchUploadGeneset,
    uberon_id: Optional[str] = None,
    output_directory: Optional[Path] = None,
    prefix: Optional[str] = None,
    gs_id: Optional[str] = None,
    hash_header: bool = False,
) -> Path:
    """Write a geneset to a CSV file.

    :param geneset: The geneset to write to a CSV file.
    :param uberon_id: The UBERON ID of the geneset.
    :param output_directory: The directory to write the CSV file to. If not provided,
    the CSV file will be written to the current working directory.
    :param prefix: A prefix to add to the CSV file name.
    :param gs_id: The ID of the geneset.
    :param hash_header: Whether to prefix the header with a hash.
    :return: The name of the CSV file.
    """
    if not gs_id:
        filename = geneset.abbreviation.replace(" ", "_").lower() + ".csv"
    else:
        filename = gs_id + ".csv"

    if prefix:
        filename = prefix + "_" + filename

    filename = Path(filename)

    header_prefix = "#" if hash_header else ""

    output_path = output_directory / filename if output_directory else filename
    header = [
        (f"{header_prefix}{key}", value)
        for key, value in geneset.dict(exclude={"values"}).items()
    ]
    header.append((f"{header_prefix}uberon_id", uberon_id))
    geneset_values = [
        (value.symbol, value.value) for value in geneset.values  # noqa: PD011
    ]

    # Write headers to a CSV file
    with open(output_path, "w", newline="") as file:
        writer = csv.writer(file)
        writer.writerows(header)
        writer.writerow(("gene_id", "value"))
        writer.writerows(geneset_values)

    return filename
